package webboard;

import bot.BotManager;
import com.corundumstudio.socketio.AckRequest;
import com.corundumstudio.socketio.Configuration;
import com.corundumstudio.socketio.SocketIOClient;
import com.corundumstudio.socketio.SocketIOServer;
import com.corundumstudio.socketio.listener.DataListener;
import com.google.gson.Gson;
import misc.Utility;
import music.BotAudioPlayer;
import music.PlayerManager;
import net.dv8tion.jda.core.entities.VoiceChannel;
import webboard.gson.BotPlayer;
import webboard.gson.SubmitToken;

import java.sql.SQLException;
import java.util.HashMap;
import java.util.Properties;
import java.util.Timer;
import java.util.TimerTask;

public class WebManager {
    private static WebManager singleton;
    private Timer timeoutDelete;
    private final HashMap<String, Session> sessionMap;
    private final HashMap<String, AuthenticatedSession> authSessionMap;
    private final SocketIOServer SERVER;
    private Properties webProperties;

    private WebManager(){
        sessionMap = new HashMap<>();
        authSessionMap = new HashMap<>();

        webProperties = new Properties();
        Utility.loadProperties("web.properties", webProperties);
        Configuration config = new Configuration();
        config.setHostname(webProperties.getProperty("socketiohost"));
        config.setPort(Integer.valueOf(webProperties.getProperty("socketioport")));
        SERVER = new SocketIOServer(config);

        addListener();

        SERVER.start();

        timeoutDelete = new Timer(true);
        TimerTask delete = new DeleteTimeoutetSessions(sessionMap, authSessionMap);
        timeoutDelete.scheduleAtFixedRate(delete, 0, 60000);
    }

    private void addListener(){
        //SERVER.addConnectListener(new ConnectListener() {
        //    @Override
        //    public void onConnect(SocketIOClient socketIOClient) {
        //        System.out.println("Someone connected");
        //    }
        //});
        SERVER.addEventListener("auth", String.class, new DataListener<String>() {
            @Override
            public void onData(SocketIOClient client, String data, AckRequest ackRequest) {
                long userID = auth(data, client);
                if(userID == -1) client.sendEvent("errorauth", "Wrong Credentials");
                else client.sendEvent("successauth", userID);
            }
        });
        SERVER.addEventListener("getsession", String.class, new DataListener<String>(){
            @Override
            public void onData(SocketIOClient client, String data, AckRequest ackRequest){
                String uuid = java.util.UUID.randomUUID().toString();
                sessionMap.put(uuid, new Session(client));
                client.sendEvent("uuid", uuid);
            }
        });
        SERVER.addEventListener("getauthurl", String.class, new DataListener<String>(){
            @Override
            public void onData(SocketIOClient client, String data, AckRequest ackRequest) {
                Session session = sessionMap.get(data);
                if(session == null) client.sendEvent("error", "No Session found");
                else client.sendEvent("authurl", session.getAuthURL());
            }
        });
        SERVER.addEventListener("submittoken", String.class, new DataListener<String>(){
            @Override
            public void onData(SocketIOClient client, String data, AckRequest ackRequest) {
                Gson gson = new Gson();
                SubmitToken token = gson.fromJson(data, SubmitToken.class);

                Session session = sessionMap.get(token.getUuid());
                if(session == null) {
                    client.sendEvent("error", "No Session found");
                    return;
                }

                session.setToken(token.getToken());
                long userID = session.getUserID();
                if(userID == -1){
                    client.sendEvent("error", "Error Retrieving UserID");
                    return;
                }
                try{
                    new SQLConnector().AddToken(userID, token.getUuid());
                    authSessionMap.put(session.getClient().getSessionId().toString(), new AuthenticatedSession(userID));
                }catch (SQLException e){
                    client.sendEvent("error", "Error Saving userID; Probably Duplicate Uuid");
                    return;
                }

                String id = String.valueOf(userID);
                client.sendEvent("userid", id);
                session.getClient().sendEvent("successauth", id);
            }
        });
        SERVER.addEventListener("player", String.class, new DataListener<String>() {
            @Override
            public void onData(SocketIOClient client, String data, AckRequest ackRequest) {
                long userID = auth(data, client);
                if(userID == -1) {
                    client.sendEvent("error", "You are not Authenticated");
                    return;
                }
                VoiceChannel channel = BotManager.getInstance().getVoiceChannel(userID);
                if(channel == null){
                    client.sendEvent("error", "You are in no Voicechannel");
                    return;
                }
                BotAudioPlayer player = PlayerManager.getInstance().getAudioPlayer(channel);
                if(player == null){
                    client.sendEvent("error", "There is no Bot in your Voicechannel");
                }
                client.sendEvent("player", new Gson().toJson(new BotPlayer(player)));
            }
        });
    }

    private long auth(String uuid, SocketIOClient client){
        if(client != null){
            AuthenticatedSession session = authSessionMap.get(client.getSessionId().toString());
            if(session != null) return session.USER_ID;
        }
        try {
            String userID = new SQLConnector().GetIDToToken(uuid);
            if(userID == null) return -1;

            long userIDLong = Long.valueOf(userID);
            if(client != null)
                authSessionMap.put(client.getSessionId().toString(), new AuthenticatedSession(userIDLong));

            return userIDLong;
        }catch (SQLException e){
            return -1;
        }
    }

    private void timeoutDeleteSessions(){

    }

    public static WebManager getInstance(){
        if(singleton != null) return singleton;

        return createSingleton();
    }

    private static synchronized WebManager createSingleton(){
        if(singleton == null) singleton = new WebManager();
        return singleton;
    }

    public static void main(String args[]){
        WebManager w = WebManager.getInstance();

        try{
            Thread.sleep(Integer.MAX_VALUE);
        }catch (Exception e){
            e.printStackTrace();
        }
    }
}
